/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.common.binaryfile;

import java.io.IOException;
import java.io.InputStream;
import java.util.Objects;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.BiFunction;

import org.apache.commons.io.IOUtils;

import de.ozgcloud.common.errorhandling.TechnicalException;
import io.grpc.stub.CallStreamObserver;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

/*
 * Q = Request Type; S = Response Type
 */
@Log4j2
public abstract class StreamingFileSender<Q, S> {

	static final int CHUNK_SIZE = 4 * 1024;
	
	private final BiFunction<byte[], Integer, Q> chunkBuilder;

	@Getter
	private final CompletableFuture<S> resultFuture = new CompletableFuture<>();

	private Q metaData;
	private final AtomicBoolean metaDataSent = new AtomicBoolean(false);
	private final AtomicBoolean done = new AtomicBoolean(false);

	@Getter(AccessLevel.PROTECTED)
	private final StreamReader streamReader;

	StreamingFileSender(BiFunction<byte[], Integer, Q> chunkBuilder, InputStream inputStream) {
		this.chunkBuilder = chunkBuilder;
		this.streamReader = new StreamReader(inputStream);
	}

	public StreamingFileSender<Q, S> withMetaData(@NonNull Q metaData) {
		this.metaData = metaData;
		return this;
	}

	public abstract StreamingFileSender<Q, S> send();

	public void cancelOnTimeout() {
		LOG.warn("File transfer canceled on timeout");
		resultFuture.cancel(true);
		getRequestObserver().onError(new TechnicalException("Timeout on waiting for upload."));
		closeStreamReader();
	}

	public void cancelOnError(Throwable t) {
		LOG.error("File transfer canceled on error.", t);
		resultFuture.cancel(true);
		getRequestObserver().onError(t);
		closeStreamReader();
	}

	void sendNext() {
		if (notFinished()) {
			waitForObserver();
			sendMetaData();
			do {
				LOG.debug("Sending next chunk.");
				sendNextChunk();
			} while (notFinished() && isReady());
			LOG.debug("Finished or waiting to become ready.");
		}
		checkIfFinishedForcefully();
	}

	private boolean notFinished() {
		return !done.get() && !resultFuture.isCancelled();
	}

	private void checkIfFinishedForcefully() {
		if (resultFuture.isCancelled()) {
			LOG.warn("File transfer was cancelled");
			closeStreamReader();
		}
	}

	private boolean isReady() {
		return getRequestObserver().isReady();
	}

	private void waitForObserver() {
		synchronized (this) {
			while (Objects.isNull(getRequestObserver())) {
				try {
					LOG.debug("wait for observer");
					wait(300);
				} catch (InterruptedException e) {
					LOG.error("Error on waiting for request Observer.", e);
					Thread.currentThread().interrupt();
				}
			}
		}

	}

	void sendNextChunk() {
		byte[] contentToSend = streamReader.getNextData();

		if (streamReader.getLastReadSize() > 0) {
			sendChunk(contentToSend, streamReader.getLastReadSize());
		} else {
			endTransfer();
		}
	}

	void endTransfer() {
		communicateEndOfTransfer();
		done.set(true);
		LOG.debug("File Transfer done.");
		closeStreamReader();
	}

	abstract void communicateEndOfTransfer();

	void closeStreamReader() {
		LOG.debug("Closing streams");
		streamReader.close();
	}

	void sendChunk(byte[] content, int length) {
		LOG.debug("Sending {} byte Data.", length);
		var chunk = chunkBuilder.apply(content, length);
		getRequestObserver().onNext(chunk);
	}

	void sendMetaData() {
		if (metaData != null && !metaDataSent.get()) {
			doSendMetaData(metaData);
		}
	}

	private void doSendMetaData(Q metadata) {
		LOG.debug("Sending Metadata.");
		getRequestObserver().onNext(metadata);
		metaDataSent.set(true);
	}

	protected abstract CallStreamObserver<Q> getRequestObserver();

	@RequiredArgsConstructor
	protected static class StreamReader {
		private final InputStream inStream;
		private final byte[] buffer = new byte[CHUNK_SIZE];
		@Getter
		private int lastReadSize = 0;
		@Getter
		private final AtomicBoolean done = new AtomicBoolean(false);

		byte[] getNextData() {
			readNext();
			return buffer;
		}

		void close() {
			IOUtils.closeQuietly(inStream);
		}

		void readNext() {
			try {
				lastReadSize = inStream.read(buffer, 0, CHUNK_SIZE);
			} catch (IOException e) {
				throw new TechnicalException("Error on reading a single chunk", e);
			}
		}
	}
}
