/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.common.binaryfile;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.BiFunction;

import org.apache.commons.lang3.RandomUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.common.binaryfile.BinaryFileTestFactory.TestRequestType;
import de.ozgcloud.common.binaryfile.BinaryFileTestFactory.TestResponseType;
import de.ozgcloud.common.errorhandling.TechnicalException;
import io.grpc.stub.CallStreamObserver;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.SneakyThrows;

class StreamingFileSenderTest {

	@Mock
	private BiFunction<byte[], Integer, TestRequestType> chunkBuilder;
	@Mock
	private InputStream inputStream;
	@Mock
	private CallStreamObserver<TestRequestType> requestObserver;

	private TestFileSender fileSender;

	@BeforeEach
	void init() {
		fileSender = spy(new TestFileSender(chunkBuilder, inputStream, requestObserver));
	}

	@Nested
	class TestCancelOnTimeout {

		@Mock
		private CompletableFuture<TestResponseType> resultFuture;

		@BeforeEach
		void init() {
			setResultFutureInFileSender(resultFuture);
		}

		@Test
		void shouldCancelResultFuture() {
			fileSender.cancelOnTimeout();

			verify(resultFuture).cancel(true);
		}

		@Test
		void shouldCallOnError() {
			fileSender.cancelOnTimeout();

			verify(requestObserver).onError(any(TechnicalException.class));
		}

		@Test
		void shouldCloseStreams() {
			fileSender.cancelOnTimeout();

			verify(fileSender).closeStreamReader();
		}
	}

	@Nested
	class TestCancelOnError {

		@Mock
		private CompletableFuture<TestResponseType> resultFuture;
		private final Throwable error = new Throwable();

		@BeforeEach
		void init() {
			setResultFutureInFileSender(resultFuture);
		}

		@Test
		void shouldCancelResultFuture() {
			fileSender.cancelOnError(error);

			verify(resultFuture).cancel(true);
		}

		@Test
		void shouldCallOnError() {
			fileSender.cancelOnError(error);

			verify(requestObserver).onError(error);
		}

		@Test
		void shouldCloseStreams() {
			fileSender.cancelOnError(error);

			verify(fileSender).closeStreamReader();
		}
	}

	@Nested
	class TestSendNext {

		@Mock
		private CompletableFuture<TestResponseType> resultFuture;

		@BeforeEach
		void init() {
			fileSender.send();
		}

		@Test
		void shouldSendMetaData() {
			fileSender.sendNext();

			verify(fileSender).sendMetaData();
		}

		@Test
		void shouldSendNextChunk() {
			fileSender.sendNext();

			verify(fileSender).sendNextChunk();
		}

		@Test
		void shouldNotSendMetaDataIfDone() {
			setDoneInFileSender(true);

			fileSender.sendNext();

			verify(fileSender, never()).sendMetaData();
		}

		@Test
		void shouldNotSendMetaDataIfCancelled() {
			fileSender.getResultFuture().cancel(true);

			fileSender.sendNext();

			verify(fileSender, never()).sendMetaData();
		}

		@Test
		void shouldSendNextChunkUntilDone() {
			lenient().when(requestObserver.isReady()).thenReturn(true);
			doAnswer(invocation -> {
				setDoneInFileSender(true);
				return null;
			}).when(fileSender).sendNextChunk();

			fileSender.sendNext();

			verify(fileSender, times(1)).sendNextChunk();
		}

		@Test
		void shouldSendNextChunkUntilCancelled() {
			lenient().when(requestObserver.isReady()).thenReturn(true);
			doAnswer(invocation -> {
				fileSender.getResultFuture().cancel(true);
				return null;
			}).when(fileSender).sendNextChunk();

			fileSender.sendNext();

			verify(fileSender, times(1)).sendNextChunk();
		}

		@Test
		void closeStreamReaderIfCancelled() {
			fileSender.getResultFuture().cancel(true);

			fileSender.sendNext();

			verify(fileSender).closeStreamReader();
		}
	}

	@Nested
	class TestSendNextChunk {

		private final byte[] content = RandomUtils.insecure().randomBytes(StreamingFileSender.CHUNK_SIZE / 2);
		private final ByteArrayInputStream byteArrayInputStream = new ByteArrayInputStream(content);

		@Captor
		private ArgumentCaptor<byte[]> chunkCaptor;

		@Nested
		class TestOnDataAvailable {

			@BeforeEach
			void init() {
				fileSender = spy(new TestFileSender(chunkBuilder, byteArrayInputStream, requestObserver));
			}

			@Test
			void shouldSendChunk() {
				fileSender.sendNextChunk();

				verify(fileSender).sendChunk(chunkCaptor.capture(), eq(content.length));
				assertThat(chunkCaptor.getValue()).contains(content);
			}
		}

		@Nested
		class TestOnNoBytesLeftToRead {

			@BeforeEach
			void init() {
				var buffer = new byte[StreamingFileSender.CHUNK_SIZE];
				byteArrayInputStream.read(buffer, 0, StreamingFileSender.CHUNK_SIZE);
				fileSender = spy(new TestFileSender(chunkBuilder, byteArrayInputStream, requestObserver));
			}

			@Test
			void shouldNotSendChunk() {
				fileSender.sendNextChunk();

				verify(fileSender, never()).sendChunk(any(), anyInt());
			}

			@Test
			void shouldEndTransfer() {
				fileSender.sendNextChunk();

				verify(fileSender).endTransfer();
			}
		}
	}

	@Nested
	class TestEndTransfer {

		@Test
		void shouldCommunicateEndOfTransfer() {
			fileSender.endTransfer();

			verify(fileSender).communicateEndOfTransfer();
		}

		@Test
		void shouldSetDoneToTrue() {
			fileSender.endTransfer();

			fileSender.sendNext();
			verify(fileSender, never()).sendNextChunk();
		}

		@Test
		void shouldCloseStreamReader() {
			fileSender.endTransfer();

			verify(fileSender).closeStreamReader();
		}
	}

	@Nested
	class TestCloseStreamReader {

		@Test
		@SneakyThrows
		void shouldCloseInputStream() {
			fileSender.closeStreamReader();

			verify(inputStream).close();
		}
	}

	@Nested
	class TestSendChunk {

		private static final byte[] CHUNK_PART = "ChunkPartContent".getBytes();

		@BeforeEach
		void initObserver() {
			fileSender.send();
		}

		@Test
		void shouldApplyBuildChunk() {
			fileSender.sendChunk(CHUNK_PART, 5);

			verify(chunkBuilder).apply(CHUNK_PART, 5);
		}

		@Test
		void shouldCallOnNext() {
			fileSender.sendChunk(CHUNK_PART, 5);

			verify(requestObserver).onNext(any());
		}
	}

	@Nested
	class TestSendMetaData {

		@Mock
		private TestRequestType metaData;

		@BeforeEach
		void initObserver() {
			fileSender.send();
		}

		@Test
		void shouldNotSendWithoutMetadata() {
			fileSender.sendMetaData();

			verify(requestObserver, never()).onNext(any());
		}

		@Test
		void shouldSendMetadata() {
			fileSender.withMetaData(metaData).sendMetaData();

			verify(requestObserver).onNext(metaData);
		}

		@Test
		void shouldSendMetadataOnlyOnce() {
			fileSender.withMetaData(metaData).sendMetaData();

			fileSender.sendMetaData();

			verify(requestObserver).onNext(metaData);
		}
	}

	private void setResultFutureInFileSender(CompletableFuture<TestResponseType> resultFuture) {
		ReflectionTestUtils.setField(fileSender, "resultFuture", resultFuture);
	}

	private void setDoneInFileSender(boolean done) {
		((AtomicBoolean) ReflectionTestUtils.getField(fileSender, null, "done")).set(done);
	}

	static class TestFileSender extends StreamingFileSender<TestRequestType, TestResponseType> {

		@Getter(AccessLevel.PROTECTED)
		private final CallStreamObserver<TestRequestType> requestObserver;

		TestFileSender(BiFunction<byte[], Integer, TestRequestType> chunkBuilder, InputStream inputStream,
				CallStreamObserver<TestRequestType> requestObserver) {
			super(chunkBuilder, inputStream);
			this.requestObserver = requestObserver;
		}

		@Override
		public StreamingFileSender<TestRequestType, TestResponseType> send() {
			return this;
		}

		@Override
		void communicateEndOfTransfer() {
		}
	}
}
